from typing import Dict, List, Optional

from pydantic import BaseModel, EmailStr, Extra, Field, root_validator, validator

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.gcp_components import GPU, Disk, VMImage
from wanna.core.utils import validators


class BucketMount(BaseModel, extra=Extra.forbid):
    bucket_name: str
    mount_path: str = "/gcs"

    _bucket_name = validator("bucket_name")(validators.validate_bucket_name)


class NotebookEnvironment(BaseModel, extra=Extra.forbid):
    vm_image: Optional[VMImage] = None
    docker_image_ref: Optional[str] = None

    _ = root_validator()(validators.validate_only_one_must_be_set)


class NotebookModel(BaseInstanceModel):
    """
    - `name`- [str] Custom name for this instance
    - `project_id' - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [Dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] Overrides network from the `gcp_profile` segment
    - `tags`- [Dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [str] (optional) Custom metadata to apply to this instance
    - `machine_type` - [str] (optional) GCP Compute Engine machine type
    - `environment` [NotebookEnvironment] (optional) Notebook Environment defined by a docker image reference
    - `owner` - [str] (optional) Currently supports one owner only. If not specified, all of the service
      account users of your VM instance’s service account can use the instance.
      If specified, only the owner will be able to access the notebook.
    - `gpu`- [GPU] (optional) The hardware GPU accelerator used on this instance.
    - `boot_disk` - [Disk] (optional) Boot disk configuration to attach to this instance.
    - `data_disk` - [Disk] (optional) Data disk configuration to attach to this instance.
    - `bucket_mounts` - [List[BucketMount]] (optional) List of buckets to be accessible from the notebook
    - `subnet`- [str] (optional) Subnetwork of a given network
    - `tensorboard_ref` - [str] (optional) Reference to Vertex Experimetes
    - `enable_monitoring` - [bool] (optional) Reports system health and notebook metrics to Cloud Monitoring
    - `collaborative` - [bool] (optional) Enable JupyterLab realtime collaboration
    https://jupyterlab.readthedocs.io/en/stable/user/rtc.html
    - `no_public_ip` - [bool] (optional) Public or private (default) IP address
    - `no_proxy_access` - [bool] (optional) If true, the notebook instance will not register with the proxy
    - `idle_shutdown_timeout` - [int] (optional) Time in minutes, between 10 and 1440. After this time of inactivity,
    notebook will be stopped. If the parameter is not set, we don't do anything.
    - `env_vars` - Dict[str, str] (optional) Environment variables to be propagated to the notebook
    - `backup` - [str] (optional) Name of the bucket where a data backup is copied (no 'gs://' needed in the name).
    After creation, any changes (including deletion) made to the data disk contents will be synced to the GCS location
    It’s recommended that you enable object versioning for the selected location so you can restore accidentally
    deleted or overwritten files. To prevent sync conflicts, avoid assigning the same location to multiple instances.
    Works only for non-Docker notebooks!
    """

    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    zone: str
    machine_type: str = "n1-standard-4"
    environment: NotebookEnvironment = NotebookEnvironment(vm_image=VMImage(framework="common", version="cpu"))
    owner: Optional[EmailStr]
    gpu: Optional[GPU]
    boot_disk: Optional[Disk]
    data_disk: Optional[Disk]
    bucket_mounts: Optional[List[BucketMount]]
    network: Optional[str]
    subnet: Optional[str]
    tensorboard_ref: Optional[str]
    enable_monitoring: bool = True
    collaborative: bool = False
    no_public_ip: bool = True
    no_proxy_access: bool = False
    idle_shutdown_timeout: Optional[int]
    env_vars: Optional[Dict[str, str]]
    backup: Optional[str]

    _machine_type = validator("machine_type")(validators.validate_machine_type)


class ManagedNotebookModel(BaseInstanceModel):
    """
    - `name`- [str] Custom name for this instance
    - `project_id' - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [Dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] (optional) Overrides network from the `gcp_profile` segment
    - `tags`- [Dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [str] (optional) Custom metadata to apply to this instance
    - `owner` - [str] This can be either a single user email address and that would be the only one
      able to access the notebook. Or service account and then everyone who has the iam.serviceAccounts.actAs
      permission on the specified service account will be able to connect.
    - `machine_type` - [str] (optional) GCP Compute Engine machine type
    - `gpu`- [GPU] (optional) The hardware GPU accelerator used on this instance.
    - `data_disk` - [Disk] (optional) Data disk configuration to attach to this instance.
    - `kernels` - [List[str]] (optional) Custom kernels given as links to container registry
    - `tensorboard_ref` - [str] (optional) Reference to Vertex Experimetes
    - `subnet`- [str] (optional) Subnetwork of a given network
    - `internal_ip_only` - [bool] (optional) Public or private (default) IP address
    - `idle_shutdown` - [bool] (optional) Turning off the notebook after the timeout, can be
      true (default) or false
    - `idle_shutdown_timeout` - [int] (optional) Time in minutes, between 10 and 1440, defaults to 180
    """

    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    owner: Optional[str]
    machine_type: Optional[str] = "n1-standard-4"
    gpu: Optional[GPU]
    data_disk: Optional[Disk]
    kernel_docker_image_refs: Optional[List[str]]
    tensorboard_ref: Optional[str]
    network: Optional[str]
    subnet: Optional[str]
    internal_ip_only: bool = True
    idle_shutdown: bool = True
    idle_shutdown_timeout: int = Field(ge=10, le=1440, default=180)



from typing import Generic, TypeVar

from wanna.core.deployment.models import GCPResource
from wanna.core.deployment.vertex_jobs import VertexJobsMixInVertex
from wanna.core.deployment.vertex_pipelines import VertexPipelinesMixInVertex

T = TypeVar("T", bound=GCPResource)


class VertexConnector(Generic[T], VertexPipelinesMixInVertex, VertexJobsMixInVertex):
    pass



from typing import List, Literal, Optional, Union

from pydantic import BaseModel, EmailStr, Field
from pydantic.typing import Annotated


class BaseNotificationChannel(BaseModel):
    """
    Base class for multiple GCP notification channels types

    :param str name: Friendly name for the channel
    :param Optional[str] description: Short description of the channel
    """

    name: str
    description: Optional[str]


class EmailNotificationChannel(BaseNotificationChannel):
    """
    EmailNotificationChannel GCP notification channel type

    :param str type: The type of channel that will be created
    :param List[str] emails: the emails to which the alerts will be sent
    """

    type: Literal["email"]
    emails: List[EmailStr]


class PubSubNotificationChannel(BaseNotificationChannel):
    """
    PubsubNotificationChannel GCP notification channel type

    :param str type: The type of channel that will be created
    :param List[str] topics: the pubsub topic to which the alerts will be sent
    """

    type: Literal["pubsub"]
    topics: List[str]


NotificationChannelModel = Annotated[
    Union[EmailNotificationChannel, PubSubNotificationChannel], Field(discriminator="type")
]



from pydantic import Field

from wanna.core.models.base_instance import BaseInstanceModel


class TensorboardModel(BaseInstanceModel):
    """
    - `name`- [str] Custom name for this instance
    - `project_id` - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `description` - [str] (optional) Tensorboard description to be shown in the UI
    - `labels`- [Dict[str, str]] (optional) Custom labels to apply to this instance
    """

    name: str = Field(min_length=3, max_length=128)
    region: str



from typing import List, Optional, Union

from pydantic import BaseModel, Extra, validator

from wanna.core.models.docker import DockerModel
from wanna.core.models.gcp_profile import GCPProfileModel
from wanna.core.models.notebook import ManagedNotebookModel, NotebookModel
from wanna.core.models.notification_channel import NotificationChannelModel
from wanna.core.models.pipeline import PipelineModel
from wanna.core.models.tensorboard import TensorboardModel
from wanna.core.models.training_custom_job import CustomJobModel, TrainingCustomJobModel
from wanna.core.models.wanna_project import WannaProjectModel
from wanna.core.utils.config_enricher import enrich_instance_with_gcp_settings


class WannaConfigModel(BaseModel, extra=Extra.forbid, validate_assignment=True):
    wanna_project: WannaProjectModel
    gcp_profile: GCPProfileModel
    docker: Optional[DockerModel]
    notebooks: List[NotebookModel] = []
    tensorboards: List[TensorboardModel] = []
    jobs: List[Union[CustomJobModel, TrainingCustomJobModel]] = []
    pipelines: List[PipelineModel] = []
    managed_notebooks: List[ManagedNotebookModel] = []
    notification_channels: List[NotificationChannelModel] = []

    _notebooks = validator("notebooks", pre=True, each_item=True, allow_reuse=True)(enrich_instance_with_gcp_settings)
    _managed_notebooks = validator("managed_notebooks", pre=True, each_item=True, allow_reuse=True)(
        enrich_instance_with_gcp_settings
    )
    _tensorboards = validator("tensorboards", pre=True, each_item=True, allow_reuse=True)(
        enrich_instance_with_gcp_settings
    )
    _jobs = validator("jobs", pre=True, each_item=True, allow_reuse=True)(enrich_instance_with_gcp_settings)
    _pipelines = validator("pipelines", pre=True, each_item=True, allow_reuse=True)(enrich_instance_with_gcp_settings)

    # TODO:
    # _gcp_profile = validator("gcp_profile", pre=True, allow_reuse=True)(enrich_gcp_profile_with_wanna_default_labels)

    @validator("notebooks", pre=True, each_item=True, allow_reuse=True)
    def validate_docker_images_defined(cls, values_inst, values):  # pylint: disable=no-self-argument,no-self-use
        docker_image_ref = values_inst.get("environment", {}).get("docker_image_ref")
        if docker_image_ref:
            if not values.get("docker"):
                raise ValueError(f"Docker image with name {docker_image_ref} is not defined")
            defined_images = [i.name for i in values.get("docker").images]
            if docker_image_ref not in defined_images:
                raise ValueError(f"Docker image with name {docker_image_ref} is not defined")
        return values_inst



#!/usr/bin/env python3

import click
from google.cloud import aiplatform


@click.command()
@click.option(
    "--project",
    envvar="CLOUD_ML_PROJECT_ID",
)
@click.option(
    "--location",
    envvar="CLOUD_ML_REGION",
)
@click.option(
    "--experiment-name",
    envvar="EXPERIMENT_NAME",
)
def main(
    project,
    location,
    experiment_name,
):
    aiplatform.init(project=project, location=location, experiment=experiment_name)


if __name__ == "__main__":
    main()



from abc import ABC
from threading import Thread
from typing import Generic, List, Optional, Tuple, TypeVar

import typer

from wanna.core.deployment.models import PushMode
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.utils.gcp import convert_project_id_to_project_number, get_network_info

logger = get_logger(__name__)

T = TypeVar("T", bound=BaseInstanceModel)


class BaseService(ABC, Generic[T]):
    """
    This is a base service and every other service (notebooks, jobs, pipelines,...)
    should build on top of this. It implements some basic function (create, delete)
    and utils for loading the wanna config.

    Args:
        instance_type: what are you working with (notebook, job, tensorboard) - used mainly in logging
        instance_model: pydantic instance model
    """

    def __init__(
        self,
        instance_type: str,
    ):
        self.instances: List[T] = []
        self.instance_type = instance_type

    def create(self, instance_name: str, **kwargs) -> None:
        """
        Create an instance with name "name" based on wanna-ml config.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be created.
                  Set to "all" to create everything from wanna-ml yaml configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._create_one_instance(instance, **kwargs)

    def delete(self, instance_name: str) -> None:
        """
        Delete an instance with name "name" based on wanna-ml config if exists on GCP.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be deleted.
                  Set to "all" to create all from configuration.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._delete_one_instance(instance)

    def stop(self, instance_name: str) -> None:
        """
        Stop an instance with name "name" based on wanna-ml config.

        Args:
            instance_name: The name of the only instance from wanna-ml config that should be stopped.
                  Set to "all" to delete everything from wanna-ml yaml configuration for this resource.
        """
        instances = self._filter_instances_by_name(instance_name)

        for instance in instances:
            self._stop_one_instance(instance)

    def _delete_one_instance(self, instance: T) -> None:
        """
        Abstract class. Should delete one instance based on one model (eg. delete one notebook).
        """
        raise NotImplementedError

    def _create_one_instance(self, instance: T, **kwargs) -> None:
        """
        Abstract class. Should create one instance based on one model (eg. create one notebook).
        """
        raise NotImplementedError

    def _stop_one_instance(self, instance: T) -> None:
        """
        Abstract class. Should stop one instance based on one model (eg. stop one job).
        """
        raise NotImplementedError

    def _filter_instances_by_name(self, instance_name: str) -> List[T]:
        """
        From self.instances filter only the instances with name instance_name.

        Args:
            instance_name: Name of the instance to return. Set to "all" to return all instances.

        Returns:
            instances

        """
        if instance_name == "all":
            instances = self.instances
            if not instances:
                logger.user_error(f"No {self.instance_type} can be parsed from your wanna-ml yaml config.")
        else:
            instances = [nb for nb in self.instances if nb.name == instance_name]
        if not instances:
            logger.user_error(
                f"{self.instance_type} with name {instance_name} not found in your wanna-ml yaml config.",
            )

        return instances

    def report(
        self,
        instance_name: str,
        wanna_project: str,
        wanna_resource: str,
        gcp_project: str,
        billing_id: Optional[str],
        organization_id: Optional[str],
    ) -> None:
        """
        Displays a link to the Google Cloud cost report
        """
        if not billing_id or not organization_id:
            logger.user_error("Billing and Organization IDs are needed. Please provide them in wanna.yaml")
            return None

        base_url = f"https://console.cloud.google.com/billing/{billing_id}/reports;projects={gcp_project}"

        if instance_name == "all":
            labels = f";labels=wanna_project:{wanna_project},wanna_resource:{wanna_resource}"
        elif instance_name not in [nb.name for nb in self.instances]:
            logger.user_error(
                f"{self.instance_type} with name {instance_name} not found in your wanna-ml yaml config.",
            )
            return None
        else:
            labels = f";labels=wanna_project:{wanna_project},wanna_resource:{wanna_resource},wanna_name:{instance_name}"

        link = base_url + labels + f"?organizationId={organization_id}"
        logger.user_info(f"Here is a link to your {wanna_resource} cost report:")
        logger.user_success(f"{link}")

    def _get_resource_network(
        self,
        project_id: str,
        push_mode: PushMode,
        resource_network: Optional[str],
        fallback_project_network: Optional[str],
        use_project_number: bool = True,
    ):
        resource_network = resource_network if resource_network else fallback_project_network
        if resource_network:
            result = get_network_info(resource_network)
            if result:
                # long format network found
                project_id, resource_network = result

            # In certain scenarios we can't build on GCP and have no access to GCP from within Avast build infra
            if push_mode.can_push_gcp_resources():
                if not project_id.isdigit() and use_project_number:
                    project_id = convert_project_id_to_project_number(project_id)

            return f"projects/{project_id}/global/networks/{resource_network}"
        else:
            return None

    def _get_resource_subnet(self, network: Optional[str], subnet: Optional[str], region: Optional[str]):
        if subnet:
            # Assumes the full qualified path was provided in config
            if "/" in subnet:
                return subnet
            else:
                # Get the project id from the provided network and pass it to subnet
                if network:
                    network_project_parts = network.split("/")
                    if len(network_project_parts) >= 2:
                        network_project_id = network_project_parts[1]
                    else:
                        network_project_id = network
                return f"projects/{network_project_id}/regions/{region}/subnetworks/{subnet}"
        else:
            return None

    def _return_diff(self) -> Tuple[List[T], List[T]]:
        """
        Abstract class.
        """
        raise NotImplementedError

    def sync(self, force: bool, push_mode: PushMode = PushMode.all) -> None:
        """
        1. Reads current instances where label is defined per field wanna_project.name in wanna.yaml
        2. Does a diff between what is on GCP and what is on yaml
        3. Delete the ones in GCP that are not in wanna.yaml
        4. Create the ones defined in yaml and missing in GCP
        """
        to_be_deleted, to_be_created = self._return_diff()  # pylint: disable=assignment-from-no-return

        if to_be_deleted:
            to_be_deleted_str = "\n".join(["- " + item.name for item in to_be_deleted])
            logger.user_info(f"{self.instance_type.capitalize()}s to be deleted:\n{to_be_deleted_str}")
            should_delete = True if force else typer.confirm("Are you sure you want to delete them?")
            if should_delete:
                for notebook in to_be_deleted:
                    t = Thread(target=self._delete_one_instance, args=(notebook,))
                    t.start()

        if to_be_created:
            to_be_created_str = "\n".join(["- " + item.name for item in to_be_created])
            logger.user_info(f"{self.instance_type.capitalize()}s to be created:\n{to_be_created_str}")
            should_create = True if force else typer.confirm("Are you sure you want to create them?")
            if should_create:
                for notebook in to_be_created:
                    t = Thread(target=self._create_one_instance, args=(notebook,), kwargs={push_mode: push_mode})
                    t.start()

        logger.user_info(f"{self.instance_type.capitalize()}s on GCP are in sync with wanna.yaml")



import sys
from pathlib import Path

from wanna.core.models.training_custom_job import JobModelTypeAlias

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from enum import Enum
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

from pydantic import BaseModel, Extra
from pydantic.generics import GenericModel

from wanna.core.models.cloud_scheduler import CloudSchedulerModel
from wanna.core.models.docker import DockerBuildResult
from wanna.core.models.notification_channel import NotificationChannelModel


class GCPResource(GenericModel, extra=Extra.forbid, validate_assignment=True, arbitrary_types_allowed=True):
    name: str
    project: str
    location: str
    service_account: Optional[str] = None

    def get_base_resource(self):
        return self.dict()


class NotificationChannelResource(GCPResource):
    type_: Literal["email", "pubsub"]
    config: Dict[str, str]
    labels: Dict[str, str]
    description: Optional[str]


class CloudSchedulerResource(GCPResource):
    cloud_scheduler: CloudSchedulerModel
    body: Dict[str, Any]
    labels: Dict[str, str]
    notification_channels: List[str]


class CloudFunctionResource(GCPResource):
    build_dir: Path
    resource_root: str
    resource_function_template: str
    resource_requirements_template: str
    template_vars: Dict[str, Any]
    env_params: Dict[str, str]
    labels: Dict[str, str]
    network: Optional[str] = None
    notification_channels: List[str]


class LogMetricResource(GCPResource):
    filter_: str
    description: str


class AlertPolicyResource(GCPResource):
    logging_metric_type: str
    resource_type: str
    display_name: str
    labels: Dict[str, str]
    notification_channels: List[str]


class PipelineResource(GCPResource):
    pipeline_name: str
    pipeline_bucket: str
    pipeline_root: str
    pipeline_version: str
    json_spec_path: str
    parameter_values: Dict[str, Any] = {}
    labels: Dict[str, str] = {}
    enable_caching: bool = True
    schedule: Optional[CloudSchedulerModel]
    docker_refs: List[DockerBuildResult]
    compile_env_params: Dict[str, str]
    network: Optional[str]
    notification_channels: List[NotificationChannelModel] = []
    encryption_spec_key_name: Optional[str]
    experiment: Optional[str]


# BaseCustomJobModel
JOB = TypeVar("JOB", bound=JobModelTypeAlias)  # dependency from wanna models


class JobResource(GCPResource, Generic[JOB]):
    job_payload: Dict[str, Any]
    image_refs: List[str] = []
    tensorboard: Optional[str]
    network: Optional[str]
    job_config: JOB
    encryption_spec: Optional[str]
    environment_variables: Optional[Dict[str, str]]


class PushArtifact(BaseModel):
    name: str


class JsonArtifact(PushArtifact):
    name: str
    json_body: Dict[Any, Any]
    destination: str


class PathArtifact(PushArtifact):
    name: str
    source: str
    destination: str


class ContainerArtifact(PushArtifact):
    name: str
    tags: List[str]


class PushMode(str, Enum):
    all = "all"
    manifests = "manifests"
    containers = "containers"
    quick = "quick"

    def is_quick_mode(self) -> bool:
        return self == PushMode.manifests or self == PushMode.quick

    def can_push_containers(self) -> bool:
        return self == PushMode.all or self == PushMode.containers

    def can_push_gcp_resources(self) -> bool:
        return self == PushMode.all or self == PushMode.manifests


class PushTask(BaseModel):
    manifest_artifacts: List[PathArtifact]
    container_artifacts: List[ContainerArtifact]
    json_artifacts: List[JsonArtifact]


PushResult = List[Tuple[List[ContainerArtifact], List[PathArtifact], List[JsonArtifact]]]



import sys
from typing import Optional

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from pydantic import BaseModel, Extra, Field, validator

from wanna.core.utils import validators


class VMImage(BaseModel, extra=Extra.forbid):
    framework: str
    version: str
    os: Optional[str] = None

    # _ = root_validator()(validators.validate_vm_image)


class Disk(BaseModel, extra=Extra.forbid):
    disk_type: str
    size_gb: int = Field(ge=100, le=65535, default=100)

    _disk_type = validator("disk_type")(validators.validate_disk_type)


class GPU(BaseModel, extra=Extra.forbid):
    count: Literal[1, 2, 4, 8]
    accelerator_type: str
    install_gpu_driver: bool = True

    _accelerator_type = validator("accelerator_type")(validators.validate_accelerator_type)



from pathlib import Path

from cookiecutter.main import cookiecutter

from wanna.components import templates
from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


def apply(output_dir: Path, no_input: bool = False):
    components_dir = Path(templates.__file__).parent.resolve()
    component_template_url = f"{components_dir}/base"
    result_dir = cookiecutter(component_template_url, output_dir=output_dir, no_input=no_input)
    logger.user_success(f"Component initiated at {result_dir}")



import json
from typing import Callable, List

from wanna.core.deployment.io import IOMixin
from wanna.core.deployment.models import ContainerArtifact, JsonArtifact, PathArtifact, PushResult, PushTask
from wanna.core.loggers.wanna_logger import get_logger

logger = get_logger(__name__)


class ArtifactsPushMixin(IOMixin):
    def push_artifacts(self, docker_pusher: Callable[[List[str]], None], push_tasks: List[PushTask]) -> PushResult:
        def push_containers(container_artifacts: List[ContainerArtifact]):
            for artifact in container_artifacts:
                with logger.user_spinner(f"Pushing {artifact.name.lower()} to {artifact.tags}"):
                    docker_pusher(artifact.tags)

        def push_manifests(manifest_artifacts: List[PathArtifact]):
            for artifact in manifest_artifacts:
                with logger.user_spinner(f"Pushing {artifact.name.lower()} to {artifact.destination}"):
                    self.upload_file(artifact.source, artifact.destination)

        def push_json(artifacts: List[JsonArtifact]):
            for artifact in artifacts:
                with logger.user_spinner(f"Pushing {artifact.name.lower()} to {artifact.destination}"):
                    self.write(artifact.destination, json.dumps(artifact.json_body))

        results: PushResult = []

        for push_task in push_tasks:

            push_containers(container_artifacts=push_task.container_artifacts)
            push_manifests(manifest_artifacts=push_task.manifest_artifacts)
            push_json(artifacts=push_task.json_artifacts)
            results.append((push_task.container_artifacts, push_task.manifest_artifacts, push_task.json_artifacts))

        return results



import sys
from enum import Enum
from pathlib import Path
from typing import Dict, List, Optional, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

from pydantic import BaseModel, Extra, Field


class DockerBuildConfigModel(BaseModel, extra=Extra.forbid):
    # Docu for more info: https://gabrieldemarmiesse.github.io/python-on-whales/sub-commands/buildx/#build
    build_args: Dict[str, str] = {}
    add_hosts: Dict[str, str] = {}
    labels: Dict[str, str] = {}
    network: Optional[str]
    platforms: Optional[List[str]]
    secrets: Union[str, List[str]] = []
    ssh: Optional[str]
    target: Optional[str]


class ImageBuildType(str, Enum):
    local_build_image = "local_build_image"
    provided_image = "provided_image"
    notebook_ready_image = "notebook_ready_image"


class BaseDockerImageModel(BaseModel, extra=Extra.forbid, validate_assignment=True):
    name: str = Field(min_length=3, max_length=128)


class LocalBuildImageModel(BaseDockerImageModel):
    """
    - `build_type` - [str] "local_build_image"
    - `name` - [str] This will later be used in `docker_image_ref` in other resources
    - `build_args` [Dict[str, str]] - (optional) docker build args
    - `context_dir` [Path] - Path to the docker build context directory
    - `dockerfile` [Path] - Path to the Dockerfile
    """

    build_type: Literal[ImageBuildType.local_build_image]
    build_args: Optional[Dict[str, str]]
    context_dir: Path
    dockerfile: Path


class ProvidedImageModel(BaseDockerImageModel):
    """
    - `build_type` - [str] "provided_image"
    - `name` - [str] This will later be used in `docker_image_ref` in other resources
    - `image_url` - [str] URL link to the image
    """

    build_type: Literal[ImageBuildType.provided_image]
    image_url: str


class NotebookReadyImageModel(BaseDockerImageModel):
    """
    - `build_type` - [str] "notebook_ready_image"
    - `name` - [str] This will later be used in `docker_image_ref` in other resources
    - `build_args` [Dict[str, str]] - (optional) docker build args
    - `base_image` [str] - (optional) base notebook docker image, you can check
    available images https://cloud.google.com/deep-learning-vm/docs/images
      when not set, it defaults to standard base CPU notebook.
    - `requirements_txt` [Path] - Path to the `requirements.txt` file containing python packages that will be installed
    """

    build_type: Literal[ImageBuildType.notebook_ready_image]
    build_args: Optional[Dict[str, str]]
    base_image: str = "gcr.io/deeplearning-platform-release/base-cpu"
    requirements_txt: Path


DockerImageModel = Union[LocalBuildImageModel, ProvidedImageModel, NotebookReadyImageModel]


class DockerModel(BaseModel, extra=Extra.forbid, validate_assignment=True):
    """
    - `images`- [List[Union[LocalBuildImageModel, ProvidedImageModel, NotebookReadyImageModel]]] Docker images
    that will be used in wanna-ml resources
    - `repository` - [str] (optional) GCP Artifact Registry repository for pushing images
    - `registry` - [str] (optional) GCP Artifact Registry, when not set it defaults
    to `{gcp_profile.region}-docker.pkg.dev`
    - `cloud_build_timeout` - [int] `12000` how many seconds before cloud build timeout
    - `cloud_build` - [str] (optional) `false` (default) to build locally, `true` to use GCP Cloud Build
    - `cloud_build_workerpool` - [str] (optional) Name of the GCP Cloud Build workerpool if you want to use one
    - `cloud_build_workerpool_location` - [str] (optional) Location of the GCP Cloud Build workerpool. Must be specified
    if cloud_build_workerpool is set.
    """

    images: List[DockerImageModel] = []
    repository: Optional[str]
    registry: Optional[str]
    cloud_build_timeout: int = 12000
    cloud_build: bool = False
    cloud_build_workerpool: Optional[str]
    cloud_build_workerpool_location: Optional[str]


class DockerBuildResult(BaseModel, extra=Extra.forbid):
    name: str
    tags: List[str]
    build_type: ImageBuildType



import functools
import os
from typing import Optional

import gcloud_config_helper
from google import auth
from google.auth import impersonated_credentials
from google.auth.credentials import Credentials
from google.auth.exceptions import DefaultCredentialsError

from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.utils.env import gcp_access_allowed

logger = get_logger(__name__)


@functools.lru_cache(maxsize=1)
def get_credentials() -> Optional[Credentials]:
    if gcp_access_allowed:
        impersonate_account = os.getenv("WANNA_IMPERSONATE_ACCOUNT")
        target_scopes = ["https://www.googleapis.com/auth/cloud-platform"]

        try:
            if impersonate_account:
                _credentials, _ = auth.default()
                credentials = impersonated_credentials.Credentials(
                    source_credentials=_credentials,
                    target_principal=impersonate_account,
                    target_scopes=_credentials.scopes if _credentials and _credentials.scopes else target_scopes,
                    lifetime=500,
                )
                return credentials
            else:
                return None

        except DefaultCredentialsError as e:
            logger.user_info(
                f"{e}\ndefault credentials were not found you likely need to execute"
                "`gcloud auth application-default login`"
            )
            exit(1)

    return None


def get_gcloud_user() -> str:
    if gcp_access_allowed:
        credentials, project = gcloud_config_helper.default()
        return credentials.properties.get("core", {}).get("account", "unidentified")[:60]

    return "no-gcp-access-not-allowed"



from google.cloud.aiplatform import (
    CustomContainerTrainingJob,
    CustomJob,
    CustomPythonPackageTrainingJob,
    HyperparameterTuningJob,
)
from google.cloud.aiplatform import hyperparameter_tuning as hpt

from wanna.core.deployment.artifacts_push import ArtifactsPushMixin
from wanna.core.deployment.models import JobResource
from wanna.core.loggers.wanna_logger import get_logger
from wanna.core.models.training_custom_job import (
    CategoricalParameter,
    CustomJobModel,
    DiscreteParameter,
    DoubleParameter,
    HyperParamater,
    IntegerParameter,
    TrainingCustomJobModel,
)

logger = get_logger(__name__)


class VertexJobsMixInVertex(ArtifactsPushMixin):
    def _create_hyperparameter_spec(
        self,
        parameter: HyperParamater,
    ) -> hpt._ParameterSpec:
        if isinstance(parameter, IntegerParameter) and parameter.type == "integer":
            return hpt.IntegerParameterSpec(min=parameter.min, max=parameter.max, scale=parameter.scale)
        elif isinstance(parameter, DoubleParameter) and parameter.type == "double":
            return hpt.DoubleParameterSpec(min=parameter.min, max=parameter.max, scale=parameter.scale)
        elif isinstance(parameter, CategoricalParameter) and parameter.type == "categorical":
            return hpt.CategoricalParameterSpec(values=parameter.values)
        elif isinstance(parameter, DiscreteParameter) and parameter.type == "discrete":
            return hpt.DiscreteParameterSpec(values=parameter.values, scale=parameter.scale)
        else:
            raise ValueError(f"Unsupported parameter type {parameter.type}")

    def run_custom_job(self, manifest: JobResource[CustomJobModel], sync: bool) -> None:
        """
        Runs a Vertex AI custom job based on the provided manifest.
        If hp_tuning parameters are set, it starts a HyperParameter Tuning instead.

        Args:
            manifest (CustomJobManifest): The Job manifest to be executed
            sync (bool): Allows to run the job in async vs sync mode

        """
        custom_job = CustomJob(**manifest.job_payload)

        if manifest.job_config.hp_tuning:

            parameter_spec = {
                param.var_name: self._create_hyperparameter_spec(param)
                for param in manifest.job_config.hp_tuning.parameters
            }
            metric_spec_dict = dict()
            for key in manifest.job_config.hp_tuning.metrics:
                metric_spec_dict[key] = str(manifest.job_config.hp_tuning.metrics[key])

            runable = HyperparameterTuningJob(
                display_name=manifest.job_config.name,
                custom_job=custom_job,
                metric_spec=metric_spec_dict,
                parameter_spec=parameter_spec,
                max_trial_count=manifest.job_config.hp_tuning.max_trial_count,
                parallel_trial_count=manifest.job_config.hp_tuning.parallel_trial_count,
                search_algorithm=manifest.job_config.hp_tuning.search_algorithm,
                encryption_spec_key_name=manifest.encryption_spec,
            )
        else:
            runable = custom_job  # type: ignore

        runable.run(
            timeout=manifest.job_config.timeout_seconds,
            enable_web_access=manifest.job_config.enable_web_access,
            tensorboard=manifest.tensorboard if manifest.tensorboard else None,
            network=manifest.network,
            sync=False,
        )

        runable.wait_for_resource_creation()
        runable_id = runable.resource_name.split("/")[-1]

        if sync:
            with logger.user_spinner(f"Running job {manifest.job_config.name} in sync mode"):
                logger.user_info(
                    f"Job Dashboard in "
                    f"https://console.cloud.google.com/vertex-ai/locations/{manifest.job_config.region}/training/{runable_id}?project={manifest.job_config.project_id}"  # noqa
                )
                custom_job.wait()
        else:
            with logger.user_spinner(f"Running job {manifest.job_config.name} in async mode"):
                logger.user_info(
                    f"Job Dashboard in "
                    f"https://console.cloud.google.com/vertex-ai/locations/{manifest.job_config.region}/training/{runable_id}?project={manifest.job_config.project_id}"  # noqa
                )

    def run_training_job(
        self,
        manifest: JobResource[TrainingCustomJobModel],
        sync: bool,
    ):
        """
        Runs a Vertex AI training custom job based on the provided manifest
        Args:
            manifest: The training Job manifest to be executed
            sync: Allows to run the job in async vs sync mode
        """

        if manifest.job_config.worker and manifest.job_config.worker.container:
            training_job = CustomContainerTrainingJob(
                training_encryption_spec_key_name=manifest.encryption_spec,
                model_encryption_spec_key_name=manifest.encryption_spec,
                **manifest.job_payload,
            )
        elif manifest.job_config.worker and manifest.job_config.worker.python_package:
            training_job = CustomPythonPackageTrainingJob(
                training_encryption_spec_key_name=manifest.encryption_spec,
                model_encryption_spec_key_name=manifest.encryption_spec,
                **manifest.job_payload,
            )  # type: ignore
        else:
            raise ValueError(
                "Wanna could not identify the type of job. Either container or python_package"
                "must be set on the owrker"
            )

        with logger.user_spinner(f"Initiating {manifest.job_config.name} custom job"):

            logger.user_info(f"Outputs will be saved to {manifest.job_config.base_output_directory}")
            training_job.run(
                machine_type=manifest.job_config.worker.machine_type,
                accelerator_type=manifest.job_config.worker.gpu.accelerator_type
                if manifest.job_config.worker.gpu and manifest.job_config.worker.gpu.accelerator_type
                else "ACCELERATOR_TYPE_UNSPECIFIED",
                accelerator_count=manifest.job_config.worker.gpu.count
                if manifest.job_config.worker.gpu and manifest.job_config.worker.gpu.count
                else 0,
                args=manifest.job_config.worker.args,
                base_output_dir=manifest.job_config.base_output_directory,
                service_account=manifest.job_config.service_account,
                network=manifest.network,
                environment_variables=manifest.job_config.worker.env,
                replica_count=manifest.job_config.worker.replica_count,
                boot_disk_type=manifest.job_config.worker.boot_disk.disk_type
                if manifest.job_config.worker.boot_disk
                else "pd-ssd",
                boot_disk_size_gb=manifest.job_config.worker.boot_disk.size_gb
                if manifest.job_config.worker.boot_disk
                else 100,
                reduction_server_replica_count=manifest.job_config.reduction_server.replica_count
                if manifest.job_config.reduction_server
                else 0,
                reduction_server_machine_type=manifest.job_config.reduction_server.machine_type
                if manifest.job_config.reduction_server
                else None,
                reduction_server_container_uri=manifest.job_config.reduction_server.container_uri
                if manifest.job_config.reduction_server
                else None,
                timeout=manifest.job_config.timeout_seconds,
                enable_web_access=manifest.job_config.enable_web_access,
                tensorboard=manifest.tensorboard if manifest.tensorboard else None,
                sync=False,
            )

        if sync:
            training_job.wait_for_resource_creation()
            job_id = training_job.resource_name.split("/")[-1]
            with logger.user_spinner(f"Running custom training job {manifest.job_config.name} in sync mode"):
                logger.user_info(
                    "Job Dashboard in "
                    f"https://console.cloud.google.com/vertex-ai/locations/{manifest.job_config.region}/training/{job_id}?project={manifest.job_config.project_id}"  # noqa
                )
                training_job.wait()
        else:
            with logger.user_spinner(f"Running custom training job {manifest.job_config.name} in async mode"):
                training_job.wait_for_resource_creation()
                job_id = training_job.resource_name.split("/")[-1]
                logger.user_info(
                    f"Job Dashboard in "
                    f"https://console.cloud.google.com/vertex-ai/locations/{manifest.job_config.region}/training/{job_id}?project={manifest.job_config.project_id}"  # noqa
                )

                # TODO:
                # Currently training_job does not release the future even with sync=False
                # and wait_for_resource_creation succeeds
                # the job is running at this stage
                # we need a "hack" to terminate main with exit 0.



from pathlib import Path

import typer

from wanna.cli.plugins.base_plugin import BasePlugin
from wanna.components import template


class ComponentsPlugin(BasePlugin):
    """
    Managing kubeflow components templated generation.
    """

    def __init__(self) -> None:
        super().__init__()
        self.register_many(
            [
                self.create,
            ]
        )

    @staticmethod
    def create(
        output_dir: Path = typer.Option(
            ...,
            "--output-dir",
            prompt="Where do you want to store the component",
            help="The output directory where wanna-ml repository will be created",
        )
    ) -> None:
        """
        Create a kubeflow component based on a template.
        """
        template.apply(output_dir)



from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, root_validator

from wanna.core.models.base_instance import BaseInstanceModel
from wanna.core.models.cloud_scheduler import CloudSchedulerModel


class PipelineModel(BaseInstanceModel):
    """
    - `name`- [str] Custom name for this instance
    - `project_id' - [str] (optional) Overrides GCP Project ID from the `gcp_profile` segment
    - `zone` - [str] (optional) Overrides zone from the `gcp_profile` segment
    - `region` - [str] (optional) Overrides region from the `gcp_profile` segment
    - `labels`- [Dict[str, str]] (optional) Custom labels to apply to this instance
    - `service_account` - [str] (optional) Overrides service account from the `gcp_profile` segment
    - `network` - [str] (optional) Overrides network from the `gcp_profile` segment
    - `tags`- [Dict[str, str]] (optional) Tags to apply to this instance
    - `metadata`- [str] (optional) Custom metadata to apply to this instance
    - `pipeline_file` - [str] Path to a Kubeflow Python script
    - `pipeline_function` - [str] (optional) Path to a cloud function
    - `pipeline_params` - [str] (optional) Path to params.yaml file
    - `docker_image_ref` - [List[str]] - List of names of docker images
    - `schedule` - [str] (optional) - Scheduler using a cron syntax
    - `tensorboard_ref` - [str] (optional) - Name of the Vertex AI Experiment
    - `notification_channels_ref` - [List[str]] (optional) List of names of notificartins channel described
    by a model: (name: str, type: Literal["email"], emails: List[EmailStr])
    - `sla_hours` - [float] (optional) Time after which the running pipeline gets stopped
    - `enable_caching` - [bool] enable KubeFlow pipeline execution caching
    """

    name: str = Field(min_length=3, max_length=63, to_lower=True, regex="^[a-z][a-z0-9-]*[a-z0-9]$")
    zone: str
    pipeline_file: Optional[str]
    pipeline_function: Optional[str]
    pipeline_params: Union[Path, Dict[str, Any], None]
    docker_image_ref: List[str] = []
    schedule: Optional[CloudSchedulerModel]
    tensorboard_ref: Optional[str]
    network: Optional[str]
    notification_channels_ref: List[str] = []
    sla_hours: Optional[float]
    enable_caching: bool = True
    experiment: Optional[str]

    @root_validator(pre=True)
    def set_experiment(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """
        Set default pipeline experiment name based on pipeline name
        """
        experiment = values.get("experiment")
        name = values.get("name")
        if not experiment and name:
            values["experiment"] = f"{name}-experiment"

        return values

    @root_validator(pre=False)
    def validate_pipeline_function(cls, values):  # pylint: disable=no-self-argument,no-self-use
        """
        Validate that at least one of pipeline_function or pipeline_file are present
        """
        pipeline_function = values.get("pipeline_function")
        pipeline_file = values.get("pipeline_file")
        if not pipeline_function and not pipeline_file:
            raise ValueError("pipeline_function or pipeline_file must be set")
        return values
